#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
Array of data parsers for bioinformatics file formats, such as: GFF3, BED, SAM/BAM, VCF, PSL, AGP, FASTA/FASTQ, BLAST, etc.
"""

from ..apps.base import dmain

if __name__ == "__main__":
    dmain(__file__)
